# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""A collection of common standard checks designed for Tanjun commands."""

from __future__ import annotations

__all__: list[str] = [
    "AuthorPermissionCheck",
    "DmCheck",
    "GuildCheck",
    "NsfwCheck",
    "OwnPermissionCheck",
    "OwnerCheck",
    "SfwCheck",
    "all_checks",
    "any_checks",
    "with_all_checks",
    "with_any_checks",
    "with_author_permission_check",
    "with_check",
    "with_dm_check",
    "with_guild_check",
    "with_nsfw_check",
    "with_own_permission_check",
    "with_owner_check",
    "with_sfw_check",
]

import typing

import alluka
import hikari

from . import _internal
from . import abc as tanjun
from . import dependencies
from . import errors
from . import permissions
from ._internal import cache
from ._internal import localisation

if typing.TYPE_CHECKING:
    from collections import abc as collections

    _ContextT_contra = typing.TypeVar("_ContextT_contra", bound=tanjun.Context, contravariant=True)

    class _AnyCallback(typing.Protocol[_ContextT_contra]):
        async def __call__(
            self, ctx: _ContextT_contra, /, *, localiser: typing.Optional[dependencies.AbstractLocaliser] = None
        ) -> bool:
            raise NotImplementedError

    _CommandT = typing.TypeVar("_CommandT", bound=tanjun.ExecutableCommand[typing.Any])
    _CallbackReturnT = typing.Union[_CommandT, collections.Callable[[_CommandT], _CommandT]]
    _MenuCommandT = typing.TypeVar("_MenuCommandT", bound=tanjun.MenuCommand[typing.Any, typing.Any])
    _MessageCommandT = typing.TypeVar("_MessageCommandT", bound=tanjun.MessageCommand[typing.Any])
    _SlashCommandT = typing.TypeVar("_SlashCommandT", bound=tanjun.BaseSlashCommand)


_ContextT = typing.TypeVar("_ContextT", bound=tanjun.Context)


def _add_to_command(command: _CommandT, check: tanjun.AnyCheckSig, follow_wrapped: bool) -> _CommandT:
    if follow_wrapped:
        for wrapped in _internal.collect_wrapped(command):
            wrapped.add_check(check)

    return command.add_check(check)


def _optional_kwargs(
    command: typing.Optional[_CommandT], check: tanjun.AnyCheckSig, follow_wrapped: bool
) -> typing.Union[_CommandT, collections.Callable[[_CommandT], _CommandT]]:
    if command:
        return _add_to_command(command, check, follow_wrapped)

    return lambda c: _add_to_command(c, check, follow_wrapped)


class _Check:
    __slots__ = ("_error", "_error_message", "_halt_execution", "_localise_id", "__weakref__")

    def __init__(
        self,
        error: typing.Optional[collections.Callable[..., Exception]],
        error_message: typing.Union[str, collections.Mapping[str, str], None],
        halt_execution: bool,
        /,
        *,
        id_name: typing.Optional[str] = None,
    ) -> None:
        self._error = error
        self._error_message = localisation.MaybeLocalised("error_message", error_message) if error_message else None
        self._halt_execution = halt_execution
        self._localise_id = f"tanjun.{id_name or type(self).__name__}"

    def _handle_result(
        self,
        ctx: tanjun.Context,
        result: bool,
        localiser: typing.Optional[dependencies.AbstractLocaliser] = None,
        /,
        *args: typing.Any,
    ) -> bool:
        if not result:
            if self._error:
                raise self._error(*args) from None
            if self._halt_execution:
                raise errors.HaltExecution from None
            if self._error_message:
                message = self._error_message.localise(ctx, localiser, "check", self._localise_id)
                raise errors.CommandError(message) from None

        return result


class OwnerCheck(_Check):
    """Standard owner check callback registered by [tanjun.with_owner_check][].

    This check will only pass if the author of the command is a bot owner.
    """

    __slots__ = ()

    def __init__(
        self,
        *,
        error: typing.Optional[collections.Callable[[], Exception]] = None,
        error_message: typing.Union[str, collections.Mapping[str, str], None] = "Only bot owners can use this command",
        halt_execution: bool = False,
    ) -> None:
        """Initialise an owner check.

        Parameters
        ----------
        error
            Callback used to create a custom error to raise if the check fails.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the
            command search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.OwnerCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        dependency: alluka.Injected[dependencies.AbstractOwners],
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        return self._handle_result(ctx, await dependency.check_ownership(ctx.client, ctx.author), localiser)


async def _get_is_nsfw(ctx: tanjun.Context, /, *, dm_default: bool) -> bool:
    if ctx.guild_id is None:
        return dm_default

    return (await cache.get_perm_channel(ctx.client, ctx.channel_id)).is_nsfw or False


class NsfwCheck(_Check):
    """Standard NSFW check callback registered by [tanjun.with_nsfw_check][].

    This check will only pass if the current channel is NSFW.
    """

    __slots__ = ()

    def __init__(
        self,
        *,
        error: typing.Optional[collections.Callable[[], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str], None
        ] = "Command can only be used in NSFW channels",
        halt_execution: bool = False,
    ) -> None:
        """Initialise a NSFW check.

        Parameters
        ----------
        error
            Callback used to create a custom error to raise if the check fails.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.NsfwCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        return self._handle_result(ctx, await _get_is_nsfw(ctx, dm_default=True), localiser)


class SfwCheck(_Check):
    """Standard SFW check callback registered by [tanjun.with_sfw_check][].

    This check will only pass if the current channel is SFW.
    """

    __slots__ = ()

    def __init__(
        self,
        *,
        error: typing.Optional[collections.Callable[[], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str], None
        ] = "Command can only be used in SFW channels",
        halt_execution: bool = False,
    ) -> None:
        """Initialise a SFW check.

        Parameters
        ----------
        error
            Callback used to create a custom error to raise if the check fails.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.SfwCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        return self._handle_result(ctx, not await _get_is_nsfw(ctx, dm_default=False), localiser)


class DmCheck(_Check):
    """Standard DM check callback registered by [tanjun.with_dm_check][].

    This check will only pass if the current channel is a DM channel.
    """

    __slots__ = ()

    def __init__(
        self,
        *,
        error: typing.Optional[collections.Callable[[], Exception]] = None,
        error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in DMs",
        halt_execution: bool = False,
    ) -> None:
        """Initialise a DM check.

        Parameters
        ----------
        error
            Callback used to create a custom error to raise if the check fails.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.DmCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)

    def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        return self._handle_result(ctx, ctx.guild_id is None, localiser)


class GuildCheck(_Check):
    """Standard guild check callback registered by [tanjun.with_guild_check][].

    This check will only pass if the current channel is in a guild.
    """

    __slots__ = ()

    def __init__(
        self,
        *,
        error: typing.Optional[collections.Callable[[], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str], None
        ] = "Command can only be used in guild channels",
        halt_execution: bool = False,
    ) -> None:
        """Initialise a guild check.

        Parameters
        ----------
        error
            Callback used to create a custom error to raise if the check fails.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.GuildCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)

    def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        return self._handle_result(ctx, ctx.guild_id is not None, localiser)


class AuthorPermissionCheck(_Check):
    """Standard author permission check callback registered by [tanjun.with_author_permission_check][].

    This check will only pass if the current author has the specified permission.
    """

    __slots__ = ("_permissions",)

    def __init__(
        self,
        permissions: typing.Union[hikari.Permissions, int],
        /,
        *,
        error: typing.Optional[collections.Callable[[hikari.Permissions], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str], None
        ] = "You don't have the permissions required to use this command",
        halt_execution: bool = False,
    ) -> None:
        """Initialise an author permission check.

        Parameters
        ----------
        permissions
            The permission(s) required for this command to run.
        error
            Callback used to create a custom error to raise if the check fails.

            This should take 1 positional argument of type [hikari.permissions.Permissions][]
            which represents the missing permissions required for this command to run.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.AuthorPermissionCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)
        self._permissions = permissions

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
    ) -> bool:
        if not ctx.member:
            # If there's no member when this is within a guild then it's likely
            # something like a webhook or guild visitor with no real permissions
            # outside of some basic set of send messages.
            if ctx.guild_id:
                perms = await permissions.fetch_everyone_permissions(ctx.client, ctx.guild_id, channel=ctx.channel_id)

            else:
                perms = permissions.DM_PERMISSIONS

        elif isinstance(ctx.member, hikari.InteractionMember):
            # Luckily, InteractionMember.permissions already handles the
            # implicit owner and admin permission special casing for us.
            perms = ctx.member.permissions

        else:
            perms = await permissions.fetch_permissions(ctx.client, ctx.member, channel=ctx.channel_id)

        missing_perms = ~perms & self._permissions
        return self._handle_result(ctx, missing_perms is hikari.Permissions.NONE, localiser, missing_perms)


_MemberCacheT = typing.Optional[dependencies.SfGuildBound[hikari.Member]]


class OwnPermissionCheck(_Check):
    """Standard own permission check callback registered by [tanjun.with_own_permission_check][].

    This check will only pass if the current bot user has the specified permission.
    """

    __slots__ = ("_permissions",)

    def __init__(
        self,
        permissions: typing.Union[hikari.Permissions, int],
        /,
        *,
        error: typing.Optional[collections.Callable[[hikari.Permissions], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str], None
        ] = "Bot doesn't have the permissions required to run this command",
        halt_execution: bool = False,
    ) -> None:
        """Initialise a own permission check.

        Parameters
        ----------
        permissions
            The permission(s) required for this command to run.
        error
            Callback used to create a custom error to raise if the check fails.

            This should take 1 positional argument of type [hikari.permissions.Permissions][]
            which represents the missing permissions required for this command to run.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            Setting this to [None][] will disable the error message allowing the command
            search to continue.

            This supports [localisation][] and uses the check name
            `"tanjun.OwnPermissionCheck"` for global overrides.
        halt_execution
            Whether this check should raise [tanjun.HaltExecution][] to
            end the execution search when it fails instead of returning [False][].

            This takes priority over `error_message`.
        """
        super().__init__(error, error_message, halt_execution)
        self._permissions = permissions

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        *,
        localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None,
        my_user: hikari.OwnUser = dependencies.inject_lc(hikari.OwnUser),
        member_cache: alluka.Injected[_MemberCacheT] = None,
    ) -> bool:
        if ctx.guild_id is None:
            perms = permissions.DM_PERMISSIONS

        elif isinstance(ctx, tanjun.AppCommandContext):
            assert ctx.interaction.app_permissions is not None
            perms = ctx.interaction.app_permissions

        elif ctx.cache and (member := ctx.cache.get_member(ctx.guild_id, my_user)):
            perms = await permissions.fetch_permissions(ctx.client, member, channel=ctx.channel_id)

        else:
            member = await member_cache.get_from_guild(ctx.guild_id, my_user.id, default=None) if member_cache else None
            member = member or await ctx.rest.fetch_member(ctx.guild_id, my_user.id)
            perms = await permissions.fetch_permissions(ctx.client, member, channel=ctx.channel_id)

        missing_perms = ~perms & self._permissions
        return self._handle_result(ctx, missing_perms is hikari.Permissions.NONE, localiser, missing_perms)


@typing.overload
def with_dm_check(command: _CommandT, /) -> _CommandT:
    ...


@typing.overload
def with_dm_check(
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in DMs",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_dm_check(
    command: typing.Optional[_CommandT] = None,
    /,
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in DMs",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> _CallbackReturnT[_CommandT]:
    """Only let a command run in a DM channel.

    Parameters
    ----------
    command : tanjun.abc.ExecutableCommand | None
        The command to add this check to.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.DmCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    tanjun.abc.ExecutableCommand
        The command this check was added to.
    """
    return _optional_kwargs(
        command, DmCheck(error=error, halt_execution=halt_execution, error_message=error_message), follow_wrapped
    )


@typing.overload
def with_guild_check(command: _CommandT, /) -> _CommandT:
    ...


@typing.overload
def with_guild_check(
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str], None
    ] = "Command can only be used in guild channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_guild_check(
    command: typing.Optional[_CommandT] = None,
    /,
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str], None
    ] = "Command can only be used in guild channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> _CallbackReturnT[_CommandT]:
    """Only let a command run in a guild channel.

    Parameters
    ----------
    command : tanjun.abc.ExecutableCommand | None
        The command to add this check to.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.GuildCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    tanjun.abc.ExecutableCommanmd
        The command this check was added to.
    """
    return _optional_kwargs(
        command, GuildCheck(error=error, halt_execution=halt_execution, error_message=error_message), follow_wrapped
    )


@typing.overload
def with_nsfw_check(command: _CommandT, /) -> _CommandT:
    ...


@typing.overload
def with_nsfw_check(
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in NSFW channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_nsfw_check(
    command: typing.Optional[_CommandT] = None,
    /,
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in NSFW channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> _CallbackReturnT[_CommandT]:
    """Only let a command run in a channel that's marked as nsfw.

    Parameters
    ----------
    command : tanjun.abc.ExecutableCommand | None
        The command to add this check to.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.NsfwCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    tanjun.abc.ExecutableCommand
        The command this check was added to.
    """
    return _optional_kwargs(
        command, NsfwCheck(error=error, halt_execution=halt_execution, error_message=error_message), follow_wrapped
    )


@typing.overload
def with_sfw_check(command: _CommandT, /) -> _CommandT:
    ...


@typing.overload
def with_sfw_check(
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in SFW channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_sfw_check(
    command: typing.Optional[_CommandT] = None,
    /,
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Command can only be used in SFW channels",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> _CallbackReturnT[_CommandT]:
    """Only let a command run in a channel that's marked as sfw.

    Parameters
    ----------
    command : tanjun.abc.ExecutableCommand | None
        The command to add this check to.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.SfwCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    tanjun.abc.ExecutableCommand
        The command this check was added to.
    """
    return _optional_kwargs(
        command, SfwCheck(error=error, halt_execution=halt_execution, error_message=error_message), follow_wrapped
    )


@typing.overload
def with_owner_check(command: _CommandT, /) -> _CommandT:
    ...


@typing.overload
def with_owner_check(
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Only bot owners can use this command",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_owner_check(
    command: typing.Optional[_CommandT] = None,
    /,
    *,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None] = "Only bot owners can use this command",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> _CallbackReturnT[_CommandT]:
    """Only let a command run if it's being triggered by one of the bot's owners.

    Parameters
    ----------
    command : tanjun.abc.ExecutableCommand | None
        The command to add this check to.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.OwnerCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    tanjun.abc.ExecutableCommand
        The command this check was added to.
    """
    return _optional_kwargs(
        command, OwnerCheck(error=error, halt_execution=halt_execution, error_message=error_message), follow_wrapped
    )


def with_author_permission_check(
    permissions: typing.Union[hikari.Permissions, int],
    *,
    error: typing.Optional[collections.Callable[[hikari.Permissions], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str], None
    ] = "You don't have the permissions required to use this command",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Only let a command run if the author has certain permissions in the current channel.

    !!! note
        This will only pass for commands in DMs if `permissions` is valid for
        a DM context (e.g. can't have any moderation permissions)

    Parameters
    ----------
    permissions
        The permission(s) required for this command to run.
    error
        Callback used to create a custom error to raise if the check fails.

        This should take 1 positional argument of type [hikari.permissions.Permissions][]
        which represents the missing permissions required for this command to run.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.AuthorPermissionCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A command decorator callback which adds the check.
    """
    return lambda command: _add_to_command(
        command,
        AuthorPermissionCheck(permissions, error=error, halt_execution=halt_execution, error_message=error_message),
        follow_wrapped,
    )


def with_own_permission_check(
    permissions: typing.Union[hikari.Permissions, int],
    *,
    error: typing.Optional[collections.Callable[[hikari.Permissions], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str], None
    ] = "Bot doesn't have the permissions required to run this command",
    follow_wrapped: bool = False,
    halt_execution: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Only let a command run if we have certain permissions in the current channel.

    !!! note
        This will only pass for commands in DMs if `permissions` is valid for
        a DM context (e.g. can't have any moderation permissions)

    Parameters
    ----------
    permissions
        The permission(s) required for this command to run.
    error
        Callback used to create a custom error to raise if the check fails.

        This should take 1 positional argument of type [hikari.permissions.Permissions][]
        which represents the missing permissions required for this command to run.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        Setting this to [None][] will disable the error message allowing the command
        search to continue.

        This supports [localisation][] and uses the check name
        `"tanjun.OwnPermissionCheck"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A command decorator callback which adds the check.
    """
    return lambda command: _add_to_command(
        command,
        OwnPermissionCheck(permissions, error=error, halt_execution=halt_execution, error_message=error_message),
        follow_wrapped,
    )


@typing.overload
def with_check(
    check: tanjun.AnyCheckSig, /, *, follow_wrapped: bool = False
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_check(
    check: tanjun.CheckSig[tanjun.MenuContext], /, *, follow_wrapped: bool = False
) -> collections.Callable[[_MenuCommandT], _MenuCommandT]:
    ...


@typing.overload
def with_check(
    check: tanjun.CheckSig[tanjun.MessageContext], /, *, follow_wrapped: bool = False
) -> collections.Callable[[_MessageCommandT], _MessageCommandT]:
    ...


@typing.overload
def with_check(
    check: tanjun.CheckSig[tanjun.SlashContext], /, *, follow_wrapped: bool = False
) -> collections.Callable[[_SlashCommandT], _SlashCommandT]:
    ...


def with_check(
    check: tanjun.CheckSig[typing.Any], /, *, follow_wrapped: bool = False
) -> typing.Union[
    collections.Callable[[_CommandT], _CommandT],
    collections.Callable[[_MenuCommandT], _MenuCommandT],
    collections.Callable[[_MessageCommandT], _MessageCommandT],
    collections.Callable[[_SlashCommandT], _SlashCommandT],
]:
    """Add a generic check to a command.

    Parameters
    ----------
    check
        The check to add to this command.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A command decorator callback which adds the check.
    """
    return lambda command: _add_to_command(command, check, follow_wrapped)


class _AllChecks(typing.Generic[_ContextT]):
    __slots__ = ("_checks", "__weakref__")

    def __init__(self, checks: list[tanjun.CheckSig[_ContextT]]) -> None:
        self._checks = checks

    async def __call__(self, ctx: _ContextT, /) -> bool:
        for check in self._checks:  # noqa: SIM111
            if not await ctx.call_with_async_di(check, ctx):
                return False

        return True


def all_checks(
    check: tanjun.CheckSig[_ContextT], /, *checks: tanjun.CheckSig[_ContextT]
) -> collections.Callable[[_ContextT], collections.Coroutine[typing.Any, typing.Any, bool]]:
    """Combine multiple check callbacks into a check which will only pass if all the callbacks pass.

    This ensures that the callbacks are run in the order they were supplied in
    rather than concurrently.

    Parameters
    ----------
    check
        The first check callback to combine.
    *checks
        Additional check callbacks to combine.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.Context], collections.abc.Coroutine[typing.Any, typing.Any, bool]]
        A check which will pass if all of the provided check callbacks pass.
    """
    return _AllChecks[_ContextT]([check, *checks])


@typing.overload
def with_all_checks(
    check: tanjun.AnyCheckSig, /, *checks: tanjun.AnyCheckSig, follow_wrapped: bool = False
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_all_checks(
    check: tanjun.CheckSig[tanjun.MenuContext],
    /,
    *checks: tanjun.CheckSig[tanjun.MenuContext],
    follow_wrapped: bool = False,
) -> collections.Callable[[_MenuCommandT], _MenuCommandT]:
    ...


@typing.overload
def with_all_checks(
    check: tanjun.CheckSig[tanjun.MessageContext],
    /,
    *checks: tanjun.CheckSig[tanjun.MessageContext],
    follow_wrapped: bool = False,
) -> collections.Callable[[_MessageCommandT], _MessageCommandT]:
    ...


@typing.overload
def with_all_checks(
    check: tanjun.CheckSig[tanjun.SlashContext],
    /,
    *checks: tanjun.CheckSig[tanjun.SlashContext],
    follow_wrapped: bool = False,
) -> collections.Callable[[_SlashCommandT], _SlashCommandT]:
    ...


def with_all_checks(
    check: tanjun.CheckSig[typing.Any], /, *checks: tanjun.CheckSig[typing.Any], follow_wrapped: bool = False
) -> typing.Union[
    collections.Callable[[_CommandT], _CommandT],
    collections.Callable[[_MenuCommandT], _MenuCommandT],
    collections.Callable[[_MessageCommandT], _MessageCommandT],
    collections.Callable[[_SlashCommandT], _SlashCommandT],
]:
    """Add a check which will pass if all the provided checks pass through a decorator call.

    This ensures that the callbacks are run in the order they were supplied in
    rather than concurrently.

    Parameters
    ----------
    check
        The first check callback to combine.
    *checks
        Additional check callbacks to combine.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.Context], collections.abc.Coroutine[typing.Any, typing.Any, bool]]
        A check which will pass if all of the provided check callbacks pass.
    """
    return lambda c: _add_to_command(c, all_checks(check, *checks), follow_wrapped)


class _AnyChecks(_Check, typing.Generic[_ContextT]):
    __slots__ = ("_checks", "_suppress")

    def __init__(
        self,
        checks: list[tanjun.CheckSig[_ContextT]],
        error: typing.Optional[collections.Callable[[], Exception]],
        error_message: typing.Union[str, collections.Mapping[str, str], None],
        halt_execution: bool,
        suppress: tuple[type[Exception], ...],
    ) -> None:
        super().__init__(error, error_message, halt_execution, id_name="any_check")
        self._checks = checks
        self._suppress = suppress

    async def __call__(
        self, ctx: _ContextT, /, *, localiser: alluka.Injected[typing.Optional[dependencies.AbstractLocaliser]] = None
    ) -> bool:
        for check in self._checks:
            try:
                if await ctx.call_with_async_di(check, ctx):
                    return True

            except errors.FailedCheck:
                pass

            except self._suppress:
                pass

        return self._handle_result(ctx, False, localiser)


def any_checks(
    check: tanjun.CheckSig[_ContextT],
    /,
    *checks: tanjun.CheckSig[_ContextT],
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> _AnyCallback[_ContextT]:
    """Combine multiple checks into a check which'll pass if any of the callbacks pass.

    This ensures that the callbacks are run in the order they were supplied in
    rather than concurrently.

    Parameters
    ----------
    check
        The first check callback to combine.
    *checks
        Additional check callbacks to combine.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        This supports [localisation][] and uses the check name
        `"tanjun.any_check"` for global overrides.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.
    suppress
        Tuple of the exceptions to suppress when a check fails.

    Returns
    -------
    collections.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A decorator which adds the generated check to a command.
    """
    return _AnyChecks[_ContextT]([check, *checks], error, error_message, halt_execution, suppress)


@typing.overload
def with_any_checks(
    check: tanjun.AnyCheckSig,
    /,
    *checks: tanjun.AnyCheckSig,
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    follow_wrapped: bool = False,
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_any_checks(
    check: tanjun.CheckSig[tanjun.MenuContext],
    /,
    *checks: tanjun.CheckSig[tanjun.MenuContext],
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    follow_wrapped: bool = False,
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> collections.Callable[[_MenuCommandT], _MenuCommandT]:
    ...


@typing.overload
def with_any_checks(
    check: tanjun.CheckSig[tanjun.MessageContext],
    /,
    *checks: tanjun.CheckSig[tanjun.MessageContext],
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    follow_wrapped: bool = False,
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> collections.Callable[[_MessageCommandT], _MessageCommandT]:
    ...


@typing.overload
def with_any_checks(
    check: tanjun.CheckSig[tanjun.SlashContext],
    /,
    *checks: tanjun.CheckSig[tanjun.SlashContext],
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    follow_wrapped: bool = False,
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> collections.Callable[[_SlashCommandT], _SlashCommandT]:
    ...


def with_any_checks(
    check: tanjun.CheckSig[typing.Any],
    /,
    *checks: tanjun.CheckSig[typing.Any],
    error: typing.Optional[collections.Callable[[], Exception]] = None,
    error_message: typing.Union[str, collections.Mapping[str, str], None],
    follow_wrapped: bool = False,
    halt_execution: bool = False,
    suppress: tuple[type[Exception], ...] = (errors.CommandError, errors.HaltExecution),
) -> typing.Union[
    collections.Callable[[_CommandT], _CommandT],
    collections.Callable[[_MenuCommandT], _MenuCommandT],
    collections.Callable[[_MessageCommandT], _MessageCommandT],
    collections.Callable[[_SlashCommandT], _SlashCommandT],
]:
    """Add a check which'll pass if any of the provided checks pass through a decorator call.

    This ensures that the callbacks are run in the order they were supplied in
    rather than concurrently.

    Parameters
    ----------
    check
        The first check callback to combine.
    *checks
        Additional check callbacks to combine.
    error
        Callback used to create a custom error to raise if the check fails.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        This supports [localisation][] and uses the check name
        `"tanjun.any_check"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    halt_execution
        Whether this check should raise [tanjun.HaltExecution][] to
        end the execution search when it fails instead of returning [False][].

        This takes priority over `error_message`.
    suppress
        Tuple of the exceptions to suppress when a check fails.

    Returns
    -------
    collections.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A decorator which adds the generated check to a command.
    """
    return lambda c: _add_to_command(
        c,
        any_checks(
            check, *checks, error=error, error_message=error_message, halt_execution=halt_execution, suppress=suppress
        ),
        follow_wrapped,
    )



# -*- coding: utf-8 -*-
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""Examples of how hooks may be used within a bot with a focus on error handling."""
import typing

import hikari

import tanjun

# While this example uses AnyHooks (which are hooks that can be used between
# both the slash command and message command flows in a similar fashion to
# checks which take `tanjun.abc.Context`), there's also `tanjun.SlashChecks`
# and `tanjun.MessageChecks` which are specialised for the relevant flows.
hooks = tanjun.AnyHooks()


@hooks.with_on_error
async def on_error(ctx: tanjun.abc.Context, exc: Exception) -> typing.Optional[bool]:
    """General error handler.

    This will be called on all errors raised during execution except errors
    derived from and including the following errors: `tanjun.CommandError`,
    `tanjun.HaltExecution` and `tanjun.ParserError`.

    To handle parser errors see `tanjun.Hooks.with_on_parser_error`.

    This must have take following positional arguments and return type:

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The command context this error was raised for.
    exc : Exception
        The error which was raised.

    Returns
    -------
    bool | None
        The return type indicates whether this hook wants the error to be
        suppressed where `True` indicates that it should be suppressed,
        `False` indicates that it should be re-raised and `None` indicates
        no decision.

        This value will be considered along with the results of any other hooks
        being called during execution by a majority rule system and if all
        hooks return `None` then the error will be raised.
    """


@hooks.with_on_parser_error
async def on_parser_error(ctx: tanjun.abc.Context, exc: tanjun.ParserError) -> None:
    """Parser error handler.

    This will be called on all parser errors raised during execution.

    It should be noted that `tanjun.Client` comes with a default parser error
    handler and calling `Client.set_hooks` will remove said handler.

    This must have the following positional arguments and return type:

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The command context this error was raised for.
    exc : Exception
        The parser error which was raised.

    Returns
    -------
    None
        Unlike general error handlers, parser errors are always suppressed if
        they're handled by a hook.
    """


@hooks.with_on_success
async def on_success(ctx: tanjun.abc.Context) -> None:
    """On success error hook.

    This will be called after a command has finished execution without raising
    any errors.

    This must have the following positional arguments and return None:

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The context of the passed command
    """


@hooks.with_post_execution
async def post_execution(ctx: tanjun.abc.Context) -> None:
    """Post-execution hook.

    This will be called after a command has finished executing regardless of
    whether it errors or passed.

    This must have the following positional arguments and return None:

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The context being executed.
    """


@hooks.with_pre_execution
async def pre_execution(ctx: tanjun.abc.Context) -> None:
    """Pre-execution hook.

    This will be called before a command has been executed.

    This must have the following positional arguments and return None:

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The context being executed.
    """


# Note that these can also be set using chained "set" methods as shown below.

hooks = (
    tanjun.AnyHooks()
    .add_on_parser_error(on_parser_error)
    .add_on_error(on_error)
    .add_on_success(on_success)
    .add_pre_execution(pre_execution)
    .add_post_execution(post_execution)
)


# AnyHooks, MessageHooks and SlashHooks may be added to commands (and command groups):


@hooks.add_to_command  # Through a decorator call
@tanjun.as_slash_command("name", "description")
async def slash_command(ctx: tanjun.abc.Context) -> None:
    ...


slash_command.set_hooks(hooks)  # or using a chainable method on the command.
# Where the hooks added to a specific command will only ever be called for said command.

# Components:


component = tanjun.Component().set_hooks(hooks)
# Where hooks set with Component.set_hooks will be called for all commands
# within the component but hooks set with `set_message_hooks` and `set_slash_hooks`
# will only be run for the message commands or slash commands within the component.


# and Clients:

bot = hikari.GatewayBot("TOKEN")
client = tanjun.Client.from_gateway_bot(bot).set_hooks(hooks)
# Where hooks set with Component.set_hooks will be called for all commands
# within the client's components but hooks set with `set_message_hooks` and `set_slash_hooks`
# will only be run for the message commands or slash commands within the client's components.



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_when_redeclare_cmds_after():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_add_modules_async_when_str_module_doesnt_exist():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_add_modules_async_when_str_module_doesnt_have_file():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_add_modules_async_when_path_module_doesnt_exist():
    ...


@pytest.mark.skip(reason="TODO")
def test_add_to_client_when_client_is_already_alive():
    ...


@pytest.mark.skip(reason="TODO")
def test_add_modules_when_str_module_doesnt_exist():
    ...


@pytest.mark.skip(reason="TODO")
def test_add_modules_when_str_module_doesnt_have_file():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_add_directory_async_when_directory_doesnt_exist():
    ...


@pytest.mark.skip(reason="TODO")
@pytest.mark.asyncio()
async def test_add_directory_async_when_targets_file():
    ...


@pytest.mark.skip(reason="TODO")
def test_add_directory_when_directory_doesnt_exist():
    ...


@pytest.mark.skip(reason="TODO")
def test_add_directory_when_targets_file():
    ...



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# This leads to too many false-positives around mocks.

from unittest import mock

import hikari
import pytest

import tanjun
from tanjun._internal import cache


@pytest.mark.asyncio()
async def test_get_perm_channel_when_hikari_cached_channel():
    mock_cache = mock.Mock()
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(543123123))

    assert result is mock_cache.get_guild_channel.return_value
    mock_cache.get_guild_channel.assert_called_once_with(543123123)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_hikari_cached_channel_and_other_caches_not_implemented():
    mock_cache = mock.Mock()
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    client = tanjun.Client(mock_rest, cache=mock_cache)

    result = await cache.get_perm_channel(client, hikari.Snowflake(543123123))

    assert result is mock_cache.get_guild_channel.return_value
    mock_cache.get_guild_channel.assert_called_once_with(543123123)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_channel():
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(34432134))

    assert result is mock_async_channel_cache.get.return_value
    mock_async_channel_cache.get.assert_called_once_with(34432134, default=None)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_channel_and_other_caches_not_implemented():
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    client = tanjun.Client(mock_rest).set_type_dependency(
        tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(34432134))

    assert result is mock_async_channel_cache.get.return_value
    mock_async_channel_cache.get.assert_called_once_with(34432134, default=None)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_channel():
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.PermissibleGuildChannel)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_TEXT
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(543345543))

    assert result is mock_rest.fetch_channel.return_value
    mock_rest.fetch_channel.assert_awaited_once_with(543345543)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_channel_and_caches_not_implemented():
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.PermissibleGuildChannel)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_TEXT
    client = tanjun.Client(mock_rest)

    result = await cache.get_perm_channel(client, hikari.Snowflake(234321))

    assert result is mock_rest.fetch_channel.return_value
    mock_rest.fetch_channel.assert_awaited_once_with(234321)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_hikari_cached_parent():
    mock_channel = mock.Mock()
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.side_effect = [None, mock_channel]
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 56431234123
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(123312))

    assert result is mock_channel
    mock_async_thread_cache.get.assert_awaited_once_with(123312, default=None)
    mock_cache.get_guild_channel.assert_has_calls([mock.call(123312), mock.call(56431234123)])


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_hikari_cached_parent_and_other_caches_not_implemented():
    mock_channel = mock.Mock()
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.side_effect = [None, mock_channel]
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 56431234123
    client = tanjun.Client(mock_rest, cache=mock_cache).set_type_dependency(
        tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(123312))

    assert result is mock_channel
    mock_async_thread_cache.get.assert_awaited_once_with(123312, default=None)
    mock_cache.get_guild_channel.assert_has_calls([mock.call(123312), mock.call(56431234123)])


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_async_cached_parent():
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 7642342
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(5454323123))

    assert result is mock_async_channel_cache.get.return_value
    mock_async_thread_cache.get.assert_awaited_once_with(5454323123, default=None)
    mock_async_channel_cache.get.assert_awaited_once_with(7642342, default=None)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_async_cached_parent_and_other_caches_not_implemented():
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = RuntimeError
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 7642342
    client = (
        tanjun.Client(mock_rest)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(5454323123))

    assert result is mock_async_channel_cache.get.return_value
    mock_async_thread_cache.get.assert_awaited_once_with(5454323123, default=None)
    mock_async_channel_cache.get.assert_awaited_once_with(7642342, default=None)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_rest_parent():
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.PermissibleGuildChannel)
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 54312343412
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(12365456))

    assert result is mock_rest.fetch_channel.return_value
    mock_async_thread_cache.get.assert_awaited_once_with(12365456, default=None)
    mock_rest.fetch_channel.assert_awaited_once_with(54312343412)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_async_cached_thread_and_rest_parent_and_other_caches_not_implemented():
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.PermissibleGuildChannel)
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value.parent_id = 54312343412
    client = tanjun.Client(mock_rest).set_type_dependency(
        tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(12365456))

    assert result is mock_rest.fetch_channel.return_value
    mock_async_thread_cache.get.assert_awaited_once_with(12365456, default=None)
    mock_rest.fetch_channel.assert_awaited_once_with(54312343412)


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_hikari_cached_parent():
    mock_channel = mock.Mock()
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.side_effect = [None, mock_channel]
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.GuildThreadChannel, parent_id=54654234)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(12312312132))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_awaited_once_with(12312312132)
    mock_cache.get_guild_channel.assert_has_calls([mock.call(12312312132), mock.call(54654234)])


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_hikari_cached_parent_and_other_caches_not_implemented():
    mock_channel = mock.Mock()
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.side_effect = [None, mock_channel]
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.GuildThreadChannel, parent_id=54654234)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    client = tanjun.Client(mock_rest, cache=mock_cache)

    result = await cache.get_perm_channel(client, hikari.Snowflake(12312312132))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_awaited_once_with(12312312132)
    mock_cache.get_guild_channel.assert_has_calls([mock.call(12312312132), mock.call(54654234)])


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_async_cached_parent():
    mock_channel = mock.Mock()
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.GuildThreadChannel, parent_id=76656345)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.side_effect = [None, mock_channel]
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(4345345))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_awaited_once_with(4345345)
    mock_async_channel_cache.get.assert_has_awaits(
        [mock.call(4345345, default=None), mock.call(76656345, default=None)]
    )


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_async_cached_parent_and_other_caches_not_implemented():
    mock_channel = mock.Mock()
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.return_value = mock.Mock(hikari.GuildThreadChannel, parent_id=76656345)
    mock_rest.fetch_channel.return_value.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.side_effect = [None, mock_channel]
    client = tanjun.Client(mock_rest).set_type_dependency(
        tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(4345345))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_awaited_once_with(4345345)
    mock_async_channel_cache.get.assert_has_awaits(
        [mock.call(4345345, default=None), mock.call(76656345, default=None)]
    )


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_rest_parent():
    mock_channel = mock.Mock(hikari.PermissibleGuildChannel)
    mock_thread = mock.Mock(hikari.GuildThreadChannel, parent_id=342234234)
    mock_thread.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    mock_cache = mock.Mock()
    mock_cache.get_guild_channel.return_value = None
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = [mock_thread, mock_channel]
    mock_async_channel_cache = mock.AsyncMock()
    mock_async_channel_cache.get.return_value = None
    mock_async_thread_cache = mock.AsyncMock()
    mock_async_thread_cache.get.return_value = None
    client = (
        tanjun.Client(mock_rest, cache=mock_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.PermissibleGuildChannel], mock_async_channel_cache)
        .set_type_dependency(tanjun.dependencies.SfCache[hikari.GuildThreadChannel], mock_async_thread_cache)
    )

    result = await cache.get_perm_channel(client, hikari.Snowflake(4543123))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_has_awaits([mock.call(4543123), mock.call(342234234)])


@pytest.mark.asyncio()
async def test_get_perm_channel_when_rest_thread_and_rest_parent_and_caches_not_implemented():
    mock_channel = mock.Mock(hikari.PermissibleGuildChannel)
    mock_thread = mock.Mock(hikari.GuildThreadChannel, parent_id=342234234)
    mock_thread.type = hikari.ChannelType.GUILD_PUBLIC_THREAD
    mock_rest = mock.AsyncMock()
    mock_rest.fetch_channel.side_effect = [mock_thread, mock_channel]
    client = tanjun.Client(mock_rest)

    result = await cache.get_perm_channel(client, hikari.Snowflake(4543123))

    assert result is mock_channel
    mock_rest.fetch_channel.assert_has_awaits([mock.call(4543123), mock.call(342234234)])



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""The errors and warnings raised within and by Tanjun."""
from __future__ import annotations

__all__: list[str] = [
    "CommandError",
    "ConversionError",
    "FailedCheck",
    "FailedModuleImport",
    "FailedModuleLoad",
    "FailedModuleUnload",
    "HaltExecution",
    "MissingDependencyError",
    "ModuleMissingLoaders",
    "ModuleMissingUnloaders",
    "ModuleStateConflict",
    "NotEnoughArgumentsError",
    "ParserError",
    "TanjunError",
    "TooManyArgumentsError",
]

import typing
from collections import abc as collections

import alluka
import hikari

if typing.TYPE_CHECKING:
    import datetime
    import pathlib

    from . import abc as tanjun


class TanjunError(Exception):
    """The base class for all errors raised by Tanjun."""


class HaltExecution(TanjunError):
    """Error raised while looking for a command in-order to end-execution early.

    For the most part, this will be raised during checks in-order to prevent
    other commands from being tried.
    """


MissingDependencyError = alluka.MissingDependencyError
"""Type alias of [alluka.MissingDependencyError][]."""


class CommandError(TanjunError):
    """An error which is sent as a response to the command call."""

    content: hikari.UndefinedOr[str]
    """The response error message's content."""

    delete_after: typing.Union[datetime.timedelta, float, int, None]
    """The seconds after which the response message should be deleted, if set."""

    attachments: hikari.UndefinedOr[collections.Sequence[hikari.Resourceish]]
    """Sequence of the attachments to be sent as a response to the command, if set."""

    components: hikari.UndefinedOr[collections.Sequence[hikari.api.ComponentBuilder]]
    """Sequence of the components to be sent as a response to the command, if set."""

    embeds: hikari.UndefinedOr[collections.Sequence[hikari.Embed]]
    """Sequence of the embeds to be sent as a response to the command, if set."""

    mentions_everyone: hikari.UndefinedOr[bool]
    """Whether or not the response should be allowed to mention `@everyone`/`@here`."""

    user_mentions: typing.Union[hikari.SnowflakeishSequence[hikari.PartialUser], bool, hikari.UndefinedType]
    """Configuration for the response's allowed user mentions.

    If this is a sequence then the response will only be allowed to mention
    users in the sequence.

    If this is a bool then the response will only be allowed to mention users
    if the value is `True`.
    """

    role_mentions: typing.Union[hikari.SnowflakeishSequence[hikari.PartialRole], bool, hikari.UndefinedType]
    """Configuration for the response's allowed role mentions.

    If this is a sequence then the response will only be allowed to mention
    roles in the sequence.

    If this is a bool then the response will only be allowed to mention roles
    if the value is `True`.
    """

    def __init__(
        self,
        content: hikari.UndefinedOr[typing.Any] = hikari.UNDEFINED,
        *,
        delete_after: typing.Union[datetime.timedelta, float, int, None] = None,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[collections.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[collections.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedOr[collections.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialUser], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
        role_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialRole], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
    ) -> None:
        """Initialise a command error.

        Parameters
        ----------
        content
            The content to respond with.

            If provided, the message contents. If
            [hikari.undefined.UNDEFINED][], then nothing will be sent
            in the content.
        delete_after
            If provided, the seconds after which the response message should be deleted.

            Slash command responses can only be deleted within 15 minutes of
            the command being received.
        attachment
            A singular attachment to respond with.
        attachments
            A sequence of attachments to respond with.
        component
            If provided, builder object of the component to include in this response.
        components
            If provided, a sequence of the component builder objects to include
            in this response.
        embed
            An embed to respond with.
        embeds
            A sequence of embeds to respond with.
        mentions_everyone
            If provided, whether the message should parse @everyone/@here
            mentions.
        user_mentions
            If provided, and [True][], all mentions will be parsed.
            If provided, and [False][], no mentions will be parsed.

            Alternatively this may be a collection of
            [hikari.snowflakes.Snowflake][], or [hikari.users.PartialUser][]
            derivatives to enforce mentioning specific users.
        role_mentions
            If provided, and [True][], all mentions will be parsed.
            If provided, and [False][], no mentions will be parsed.

            Alternatively this may be a collection of
            [hikari.snowflakes.Snowflake][], or [hikari.guilds.PartialRole][]
            derivatives to enforce mentioning specific roles.

        Raises
        ------
        ValueError
            Raised for any of the following reasons:

            * When both `attachment` and `attachments` are provided.
            * When both `component` and `components` are passed.
            * When both `embed` and `embeds` are passed.
            * If more than 100 entries are passed for `role_mentions`.
            * If more than 100 entries are passed for `user_mentions`.
        """
        if attachment and attachments:
            raise ValueError("Cannot specify both attachment and attachments")

        if component and components:
            raise ValueError("Cannot specify both component and components")

        if embed and embeds:
            raise ValueError("Cannot specify both embed and embeds")

        if isinstance(role_mentions, collections.Sequence) and len(role_mentions) > 100:
            raise ValueError("Cannot specify more than 100 role mentions")

        if isinstance(user_mentions, collections.Sequence) and len(user_mentions) > 100:
            raise ValueError("Cannot specify more than 100 user mentions")

        self.attachments = [attachment] if attachment else attachments
        self.content = content
        self.components = [component] if component else components
        self.delete_after = delete_after
        self.embeds = [embed] if embed else embeds
        self.mentions_everyone = mentions_everyone
        self.role_mentions = role_mentions
        self.user_mentions = user_mentions

    def __str__(self) -> str:
        return self.content or ""

    @typing.overload
    async def send(self, ctx: tanjun.Context, /, *, ensure_result: typing.Literal[True]) -> hikari.Message:
        ...

    @typing.overload
    async def send(self, ctx: tanjun.Context, /, *, ensure_result: bool = False) -> typing.Optional[hikari.Message]:
        ...

    async def send(self, ctx: tanjun.Context, /, *, ensure_result: bool = False) -> typing.Optional[hikari.Message]:
        """Send this error as a command response.

        Parameters
        ----------
        ctx
            The command call context to respond to.
        ensure_result
            Ensure that this call will always return a message object.

            If [True][] then this will always return [hikari.messages.Message][],
            otherwise this will return `hikari.Message | None`.

            It's worth noting that, under certain scenarios within the slash
            command flow, this may lead to an extre request being made.

        Raises
        ------
        ValueError
            If `delete_after` would be more than 15 minutes after the slash
            command was called.
        hikari.BadRequestError
            This may be raised in several discrete situations, such as messages
            being empty with no attachments or embeds; messages with more than
            2000 characters in them, embeds that exceed one of the many embed
            limits; too many attachments; attachments that are too large;
            invalid image URLs in embeds; too many components.
        hikari.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.ForbiddenError
            If you are missing the `SEND_MESSAGES` in the channel or the
            person you are trying to message has the DM's disabled.
        hikari.NotFoundError
            If the channel is not found.
        hikari.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await ctx.respond(
            content=self.content,
            attachments=self.attachments,
            components=self.components,
            delete_after=self.delete_after,
            embeds=self.embeds,
            ensure_result=ensure_result,
            mentions_everyone=self.mentions_everyone,
            role_mentions=self.role_mentions,
            user_mentions=self.user_mentions,
        )


# TODO: use this
class InvalidCheck(TanjunError, RuntimeError):  # TODO: or/and warning?  # TODO: InvalidCheckError
    """Error raised as an assertion that a check will never pass in the current environment."""


class FailedCheck(TanjunError, RuntimeError):  # TODO: FailedCheckError
    """Error raised as an alternative to returning `False` in a check."""


class ParserError(TanjunError, ValueError):
    """Base error raised by a parser or parameter during parsing.

    !!! note
        Expected errors raised by the parser will subclass this error.
    """

    message: str
    """String message for this error.

    !!! note
        This may be used as a command response message.
    """

    parameter: typing.Optional[str]
    """Name of the this was raised for.

    !!! note
        This will be [None][] if it was raised while parsing the
        provided message content.
    """

    def __init__(self, message: str, parameter: typing.Optional[str], /) -> None:
        """Initialise a parser error.

        Parameters
        ----------
        message
            String message for this error.
        parameter
            Name of the parameter which caused this error, should be [None][]
            if not applicable.
        """
        self.message = message
        self.parameter = parameter

    def __str__(self) -> str:
        return self.message


class ConversionError(ParserError):
    """Error raised by a parser parameter when it failed to converter a value."""

    errors: collections.Sequence[ValueError]
    """Sequence of the errors that were caught during conversion for this parameter."""

    parameter: str
    """Name of the parameter this error was raised for."""

    def __init__(self, message: str, parameter: str, /, *, errors: collections.Iterable[ValueError] = ()) -> None:
        """Initialise a conversion error.

        Parameters
        ----------
        parameter
            The parameter this was raised by.
        errors
            An iterable of the source value errors which were raised during conversion.
        """
        super().__init__(message, parameter)
        self.errors = tuple(errors)


class NotEnoughArgumentsError(ParserError):
    """Error raised by the parser when not enough arguments are found for a parameter."""

    parameter: str
    """Name of the parameter this error was raised for."""

    def __init__(self, message: str, parameter: str, /) -> None:
        """Initialise a not enough arguments error.

        Parameters
        ----------
        message
            The error message.
        parameter
            The parameter this error was raised for.
        """
        super().__init__(message, parameter)


class TooManyArgumentsError(ParserError):
    """Error raised by the parser when too many arguments are found for a parameter."""

    parameter: str
    """Name of the parameter this error was raised for."""

    def __init__(self, message: str, parameter: str, /) -> None:
        """Initialise a too many arguments error.

        Parameters
        ----------
        message
            The error message.
        parameter
            The parameter this error was raised for.
        """
        super().__init__(message, parameter)


class ModuleMissingLoaders(RuntimeError, TanjunError):
    """Error raised when a module is missing loaders."""

    def __init__(self, message: str, path: typing.Union[str, pathlib.Path], /) -> None:
        self._message = message
        self._path = path

    @property
    def message(self) -> str:
        """The error message."""
        return self._message

    @property
    def path(self) -> typing.Union[str, pathlib.Path]:
        """The path of the module which is missing loaders."""
        return self._path


class ModuleMissingUnloaders(RuntimeError, TanjunError):
    """Error raised when a module is missing unloaders."""

    def __init__(self, message: str, path: typing.Union[str, pathlib.Path], /) -> None:
        self._message = message
        self._path = path

    @property
    def message(self) -> str:
        """The error message."""
        return self._message

    @property
    def path(self) -> typing.Union[str, pathlib.Path]:
        """The path of the module which is missing unloaders."""
        return self._path


class ModuleStateConflict(ValueError, TanjunError):
    """Error raised when a module cannot be (un)loaded due to a state conflict."""

    def __init__(self, message: str, path: typing.Union[str, pathlib.Path], /) -> None:
        self._message = message
        self._path = path

    @property
    def message(self) -> str:
        """The error message."""
        return self._message

    @property
    def path(self) -> typing.Union[str, pathlib.Path]:
        """The path of the module which caused the error."""
        return self._path


class FailedModuleLoad(TanjunError):
    """Error raised when a module fails to load.

    This may be raised by the module failing to import or by one of
    its loaders erroring.

    This source error can be accessed at
    [FailedModuleLoad.__cause__][tanjun.errors.FailedModuleLoad.__cause__].
    """

    __cause__: Exception
    """The root error."""

    def __init__(self, path: typing.Union[str, pathlib.Path], /) -> None:
        self._path = path

    @property
    def path(self) -> typing.Union[str, pathlib.Path]:
        """The path of the module which caused the error."""
        return self._path


class FailedModuleImport(FailedModuleLoad):
    """Error raised when a module failed to import.

    This is a specialisation of [tanjun.errors.FailedModuleLoad][].
    """


class FailedModuleUnload(TanjunError):
    """Error raised when a module fails to unload.

    This may be raised by the module failing to import or by one
    of its unloaders erroring.

    The source error can be accessed at
    [FailedModuleUnload.__cause__][tanjun.errors.FailedModuleUnload.__cause__].
    """

    __cause__: Exception
    """The root error."""

    def __init__(self, path: typing.Union[str, pathlib.Path], /) -> None:
        self._path = path

    @property
    def path(self) -> typing.Union[str, pathlib.Path]:
        """The path of the module which caused the error."""
        return self._path
